#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import pathlib
from typing import Union

import numpy as np
import networkx as nx

from ase.io import read, write
from ase.neighborlist import natural_cutoffs, NeighborList

from .comparator import AbstractComparator

from ..builder.group import create_a_group


bond_match = nx.algorithms.isomorphism.categorical_edge_match("bond", "")

def create_a_graph(atoms, indices=None):
    """"""
    natoms = len(atoms)
    if indices is None:
        indices = range(natoms)

    graph = nx.Graph()
    for i in indices:
        graph.add_node(atoms.symbols[i]+"_"+str(i), )
    
    nl = NeighborList(
        cutoffs=natural_cutoffs(atoms, mult=1.), skin=0.2, sorted=False,
        self_interaction=False, bothways=False
    )
    nl.update(atoms)
    for i in indices: # TODO: PBC
        nei_indices, nei_offsets = nl.get_neighbors(i)
        s_i = atoms.symbols[i]
        for j in nei_indices:
            s_j = atoms.symbols[j]
            graph.add_edge(
                f"{s_i}_{i}", f"{s_j}_{j}",
                bond = "{}{}".format(*sorted([s_i, s_j]))
            )

    return graph


# Function to calculate the inertia tensor for a point mass
def point_mass_inertia_tensor(mass, position):
    I = np.zeros((3, 3))
    r_squared = np.dot(position, position)

    for i in range(3):
        for j in range(3):
            if i == j:
                I[i, j] = mass * (r_squared - position[i]**2)
            else:
                I[i, j] = -mass * position[i] * position[j]

    return I

# Function to calculate the total inertia tensor for the nanoparticle
def calculate_inertia_tensor(coordinates, atomic_masses):
    total_inertia_tensor = np.zeros((3, 3))

    # Iterate through each copper atom and add its contribution to the total inertia tensor
    for i in range(len(coordinates)):
        atom_position, atomic_mass = coordinates[i], atomic_masses[i]
        atom_inertia_tensor = point_mass_inertia_tensor(atomic_mass, atom_position)
        total_inertia_tensor += atom_inertia_tensor

    return total_inertia_tensor


class GraphComparator(AbstractComparator):

    dtol_avg: float = 0.1 # displacement tolerance tolerance, Ang
    dtol_std: float = 0.02 # displacement tolerance tolerance, Ang

    group: str = None

    def __init__(self, dtol_avg=0.1, dtol_std=0.02, mic=True, group=None, *args, **kwargs):
        """"""
        super().__init__(*args, **kwargs)

        self.dtol_avg = dtol_avg
        self.dtol_std = dtol_std

        self.mic = mic
        self.group = group

        return
    
    def looks_like(self, a1, a2):
        """"""
        is_similar = self.compare_composition(a1, a2)
        if is_similar:
            # -
            ainds = None
            if self.group is not None:
                g1 = create_a_group(a1, self.group)
                g2 = create_a_group(a2, self.group)
                if g1 == g2: # can be []
                    ainds = g1
                else:
                    ainds = []
                if len(ainds) > 0:
                    self._print(f"natoms: {len(ainds)}")
                    self._print(f"{a1[ainds].get_chemical_formula()}")
                    #write("xxx.xyz", a1[ainds])
                else:
                    ainds = range(len(a1)) # atomic indices
            else:
                ainds = range(len(a1)) # atomic indices
            # - 
            graph_1 = create_a_graph(a1, ainds)
            graph_2 = create_a_graph(a2, ainds)
            matcher = nx.algorithms.isomorphism.GraphMatcher(
                graph_1, graph_2, edge_match=bond_match
            )
            is_isomorphic = matcher.is_isomorphic()
            self._print(f"  isomorphic: {is_isomorphic}")
            if is_isomorphic:
                ...
                #inertia_1 = calculate_inertia_tensor(
                #    a1.positions[ainds, :], a1.get_masses()[ainds]
                #)
                #eig_1 = np.linalg.eigvals(inertia_1)
                #inertia_2 = calculate_inertia_tensor(
                #    a2.positions[ainds, :], a2.get_masses()[ainds]
                #)
                #eig_2 = np.linalg.eigvals(inertia_2)
                #self._print("  "+str(eig_1 - eig_2))
                #self._print("  "+str(np.linalg.norm(eig_1 - eig_2)))
                #if np.linalg.norm(eig_1 - eig_2) >= 40.:
                #    is_similar = False
            else:
                is_similar = False
        else:
            ...
        self._print(f" similar: {is_similar}")

        return is_similar


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import dataclasses
import pathlib
import traceback

from typing import Callable, List

import numpy as np

from ase import Atoms
from ase.io import read, write
from ase.neb import NEB, NEBTools
from ase.calculators.singlepoint import SinglePointCalculator
from ase.optimize.optimize import Dynamics


from .. import config as GDPCONFIG
from .. import EnhancedCalculator
from .string import AbstractStringReactor, StringReactorSetting


def update_atoms_info(
    neb: NEB,
    dyn: Dynamics,
    start_step: int = 0,
    print_func: Callable = print,
    debug_func: Callable = print,
) -> None:
    """"""
    print_func(f"step = {dyn.nsteps}")

    for i, atoms in enumerate(neb.iterimages()):
        atoms.info["step"] = dyn.nsteps + start_step
        atoms.info["image"] = i

    return


def convert_atoms(atoms: Atoms) -> Atoms:
    # - save atoms
    atoms_to_save = Atoms(
        symbols=atoms.get_chemical_symbols(),
        positions=atoms.get_positions().copy(),
        cell=atoms.get_cell().copy(),
        pbc=copy.deepcopy(atoms.get_pbc()),
    )
    if "tags" in atoms.arrays:
        atoms_to_save.set_tags(atoms.get_tags())
    if atoms.get_kinetic_energy() > 0.0:
        atoms_to_save.set_momenta(atoms.get_momenta())
    results = dict(
        energy=atoms.get_potential_energy(),
        forces=copy.deepcopy(atoms.get_forces()),
    )
    spc = SinglePointCalculator(atoms, **results)
    atoms_to_save.calc = spc

    # - save atoms info
    # FIXME: Why cannot get step by ["step"]???
    atoms_to_save.info["step"] = atoms.info.get("step")
    atoms_to_save.info["image"] = atoms.info.get("image")

    # - save special keys and arrays from calc
    natoms = len(atoms)
    # -- add deviation
    for k, v in atoms.calc.results.items():
        if k in GDPCONFIG.VALID_DEVI_FRAME_KEYS:
            atoms_to_save.info[k] = v
    for k, v in atoms.calc.results.items():
        if k in GDPCONFIG.VALID_DEVI_ATOMIC_KEYS:
            atoms_to_save.arrays[k] = np.reshape(v, (natoms, -1))

    # -- check special metadata
    calc = atoms.calc
    if isinstance(calc, EnhancedCalculator):
        atoms_to_save.info["host_energy"] = copy.deepcopy(calc.results["host_energy"])
        atoms_to_save.arrays["host_forces"] = copy.deepcopy(calc.results["host_forces"])

    return atoms_to_save


def save_nebtraj(neb: NEB, nebtraj_fpath: pathlib.Path) -> None:
    """Create a clean atoms from the input and save simulation trajectory.

    We need an explicit copy of atoms as some calculators may not return all
    necessary information. For example, schnet only returns required properties.
    If only energy is required, there are no forces.

    """
    # - append to traj
    for atoms in neb.iterimages():
        atoms_to_save = convert_atoms(atoms)
        write(nebtraj_fpath, atoms_to_save, append=True)

    return


@dataclasses.dataclass
class AseStringReactorSetting(StringReactorSetting):

    backend: str = "ase"

    def __post_init__(self):
        """"""
        # - ...
        opt_cls = None
        if self.optimiser == "bfgs":  # Takes a lot of time to solve hessian.
            from ase.optimize import BFGS as opt_cls
        elif self.optimiser == "fire":  # BUG: STRANGE BEHAVIOUR.
            from ase.optimize import FIRE as opt_cls
        elif self.optimiser == "mdmin":
            from ase.optimize import MDMin as opt_cls
        else:
            ...

        self.opt_cls = opt_cls

        return

    def get_run_params(self, *args, **kwargs):
        """"""
        steps_ = kwargs.get("steps", self.steps)
        if steps_ <= 0:
            steps_ = -1
        run_params = dict(steps=steps_, fmax=self.fmax)

        return run_params


class AseStringReactor(AbstractStringReactor):
    """Find the minimum energy path based on input structures.

    Methods based on the number of input structures such as single, double, multi...

    """

    name = "ase"

    traj_name: str = "nebtraj.xyz"

    def __init__(
        self,
        calc=None,
        params={},
        ignore_convergence=False,
        directory="./",
        *args,
        **kwargs,
    ) -> None:
        """"""
        self.calc = calc
        if self.calc is not None:
            self.calc.reset()

        self.ignore_convergence = ignore_convergence

        self.directory = directory
        self.cache_nebtraj = self.directory / self.traj_name

        # - parse params
        self.setting = AseStringReactorSetting(**params)
        self._debug(self.setting)

        return

    @AbstractStringReactor.directory.setter
    def directory(self, directory_):
        self._directory = pathlib.Path(directory_)
        self.calc.directory = str(self.directory)  # NOTE: avoid inconsistent in ASE

        self.cache_nebtraj = self.directory / self.traj_name

        return

    def _verify_checkpoint(self, *args, **kwargs) -> bool:
        """"""
        verified = super()._verify_checkpoint(*args, **kwargs)
        if verified:
            if self.cache_nebtraj.exists() and self.cache_nebtraj.stat().st_size != 0:
                verified = True
            else:
                verified = False
        else:
            ...

        return verified

    def _irun(self, structures: List[Atoms], ckpt_wdir=None, *args, **kwargs):
        """"""
        try:
            run_params = self.setting.get_run_params()

            start_step = 0
            if ckpt_wdir is None:  # start from the scratch
                images = self._align_structures(structures, run_params)
                write(self.directory / "images.xyz", images)
            else:
                nebtraj = self.read_trajectory()
                images = nebtraj[-1]
                start_step = images[0].info["step"]
                run_params["steps"] = run_params["steps"] - start_step

            for a in images:
                a.calc = self.calc

            neb = NEB(
                images=images,
                k=self.setting.k,
                climb=self.setting.climb,
                remove_rotation_and_translation=False,
                method="aseneb",
                allow_shared_calculator=True,
                precon=None,
            )

            dynamics = self.setting.opt_cls(
                neb,
                logfile=self.directory / "neb.log",
                trajectory=None,
            )
            dynamics.attach(
                update_atoms_info,
                interval=self.setting.dump_period,
                neb=neb,
                dyn=dynamics,
                start_step=start_step,
                print_func=self._print,
                debug_func=self._debug,
            )
            dynamics.attach(
                save_nebtraj,
                interval=self.setting.dump_period,
                neb=neb,
                nebtraj_fpath=self.cache_nebtraj,
            )

            # steps = run_params["steps"]
            # self._print(f"{steps =}")
            dynamics.run(steps=run_params["steps"], fmax=run_params["fmax"])

            # Always save the last step
            dump_period = self.setting.dump_period
            if dump_period > 1:
                # optimiser dumps every step to log!!
                data = np.loadtxt(self.directory / "neb.log", dtype=str, skiprows=1)
                if len(data.shape) == 1:
                    data = data[np.newaxis, :]
                nsteps = data.shape[0]
                if nsteps > 0 and (nsteps - 1) % dump_period != 0:
                    update_atoms_info(
                        neb,
                        dynamics,
                        start_step=start_step,
                        print_func=self._print,
                        debug_func=self._debug,
                    )
                    save_nebtraj(neb, self.cache_nebtraj)
            else:
                ...
        except Exception as e:
            self._debug(f"Exception of {self.__class__.__name__} is {e}.")
            self._debug(
                f"Exception of {self.__class__.__name__} is {traceback.format_exc()}."
            )

        return

    def _read_a_single_trajectory(self, wdir, *args, **kwargs):
        """"""
        cache_nebtraj = wdir / self.traj_name
        nimages_per_band = self.setting.nimages
        if cache_nebtraj.exists():
            images = read(cache_nebtraj, ":")
        else:
            raise FileNotFoundError(f"No cache trajectory {str(cache_nebtraj)}.")

        nimages = len(images)
        assert nimages % nimages_per_band == 0, "Inconsistent number of bands."
        nbands = int(nimages / nimages_per_band)

        reshaped_images = []
        for i in range(nbands):
            reshaped_images.append(
                images[i * nimages_per_band : (i + 1) * nimages_per_band]
            )

        return reshaped_images

    def read_convergence(self, *args, **kwargs) -> bool:
        """"""
        converged = False
        if self.cache_nebtraj.exists():
            images = self.read_trajectory()
            nsteps = images[-1][0].info["step"] + 1

            # end_band = images[-1]
            # for a in end_band:
            #     self._preprocess_constraints(a, self.setting.constraint)
            # self._print(end_band[3].get_forces(apply_constraint=True))
            # self._print(np.max(np.fabs(end_band[3].get_forces(apply_constraint=True))))

            # NOTE: Read convergece from neb.log instead of computing fmax from 
            #       structures as sometimes it is inconsistent due to some reasons
            #       (precision?)

            # Step     Time          Energy         fmax
            # BFGS:    0 21:24:07     -195.368118        5.8916
            data = np.loadtxt(self.directory / "neb.log", dtype=str, skiprows=1)
            if len(data.shape) == 1:
                data = data[np.newaxis, :]
            end_step = data.shape[0]
            assert end_step == nsteps, "Inconsistent `end_step` in neb.log and nebtraj.xyz."
            end_fmax = float(data[-1, -1])

            # nt = NEBTools(end_band)
            # fmax = nt.get_fmax()

            if (end_fmax <= self.setting.fmax) or (end_step >= self.setting.steps + 1):
                converged = True
            self._print(
                f"STEP: {end_step} >= {self.setting.steps} MAXFRC: {end_fmax} <=? {self.setting.fmax}"
            )

        return converged


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import itertools
from typing import NoReturn, List
import yaml

import numpy as np

from ase import Atoms
from ase import data, units
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution
from ase.neighborlist import NeighborList, natural_cutoffs

from .. import convert_string_to_atoms
from .operator import AbstractOperator


class BasicExchangeOperator(AbstractOperator):

    MIN_RANDOM_TAG: int = 10000

    MAX_RANDOM_TAG: int = 100000

    def _insert(self, atoms_, species: str, rng=np.random.Generator(np.random.PCG64())):
        """"""
        atoms = atoms_.copy()

        # - prepare particle to add
        adpart = convert_string_to_atoms(species)

        # - add velocity in case the mixed MC/MD is performed
        MaxwellBoltzmannDistribution(adpart, temperature_K=self.temperature, rng=rng)

        # - add tag
        used_tags = set(atoms.get_tags().tolist())
        adpart_tag = 0
        while adpart_tag in used_tags:
            # NOTE: np.random only has randint
            adpart_tag = rng.integers(self.MIN_RANDOM_TAG, self.MAX_RANDOM_TAG)
        adpart_tag = int(adpart_tag)
        self._print(
            f"adpart {adpart.get_chemical_formula()} tag: {adpart_tag} {type(adpart_tag)}"
        )
        # NOTE: ase accepts int or list as tags
        adpart.set_tags(adpart_tag)

        # - add particle
        atoms.extend(adpart)

        # - find species indices
        #   NOTE: adpart is always added to the end
        species_indices = [i for i, t in enumerate(atoms.get_tags()) if t == adpart_tag]

        # - blmin is initialised by MC
        cell = atoms.get_cell(complete=True)

        # - neighbour list
        nl = NeighborList(
            self.covalent_max * np.array(natural_cutoffs(atoms)),
            skin=0.0,
            self_interaction=False,
            bothways=True,
        )

        # - get a random position
        species = atoms[species_indices]
        org_com = np.mean(species.positions, axis=0)
        org_positions = species.positions.copy()

        for i in range(self.MAX_RANDOM_ATTEMPTS):
            # - make a copy
            species_ = self._rotate_species(species, rng=rng)
            curr_cop = np.average(species_.positions, axis=0)
            ran_pos = self.region.get_random_positions(size=1, rng=rng)[0]
            new_vec = ran_pos - curr_cop
            species_.translate(new_vec)
            atoms.positions[species_indices] = copy.deepcopy(species_.positions)
            if not self.check_overlap_neighbour(nl, atoms, cell, species_indices):
                self._print(f"succeed to insert after {i+1} attempts...")
                self._print(f"original position: {org_com}")
                self._print(f"random position: {ran_pos}")
                self._print(
                    f"actual position: {np.average(atoms.positions[species_indices], axis=0)}"
                )
                break
            atoms.positions[species_indices] = org_positions
        else:
            # -- remove adpart since the insertion fails
            del atoms[species_indices]
            atoms = None

        return atoms

    def _remove(self, atoms_, species: str, rng=np.random.Generator(np.random.PCG64())):
        """"""
        atoms = atoms_.copy()

        # - pick a random atom/molecule
        species_indices = self._select_species(atoms, [species], rng)

        # - remove then
        del atoms[species_indices]

        return atoms


class ExchangeOperator(BasicExchangeOperator):

    name: str = "exchange"

    #: The current suboperation (insert or remove).
    _curr_operation: str = None  # insert or remove

    #: The current tags dict.
    _curr_tags_dict: dict = None

    #: The current accpetable volume.
    _curr_volume: float = None

    def __init__(
        self,
        region: dict,
        reservoir: dict,
        temperature: float = 300.0,
        pressure: float = 1.0,
        covalent_ratio=[0.8, 2.0],
        use_rotation: bool = True,
        use_bias: bool = True,
        *args,
        **kwargs,
    ):
        """"""
        super().__init__(
            region=region,
            temperature=temperature,
            pressure=pressure,
            covalent_ratio=covalent_ratio,
            use_rotation=use_rotation,
            *args,
            **kwargs,
        )

        self.species = reservoir["species"]
        self.mu = reservoir["mu"]

        self.use_bias = use_bias

        return

    def run(self, atoms: Atoms, rng=np.random) -> Atoms:
        """"""
        super().run(atoms)
        self._extra_info = "-"

        # -- compute acceptable volume
        if self.use_bias:
            # --- determine on-the-fly
            acc_volume = self.region.get_empty_volume(atoms)
        else:
            # --- get normal region
            acc_volume = self.region.get_volume()
        self._curr_volume = acc_volume

        # - choose a species to exchange
        # valid_species = [k for k, v in tag_dict.items() if len(v) > 0]
        nparticles = len(self._curr_tags_dict.get(self.species, []))

        # - choose insert or remove
        if nparticles > 0:
            rn_ex = rng.uniform()
            if rn_ex < 0.5:
                self._print("...insert...")
                self._curr_operation = "insert"
                curr_atoms = self._insert(atoms, self.species, rng)
                self._extra_info = f"Insert_{self.species}"
            else:
                self._print("...remove...")
                self._curr_operation = "remove"
                curr_atoms = self._remove(atoms, self.species, rng)
                self._extra_info = f"Remove_{self.species}"
        else:
            self._print("...insert...")
            self._curr_operation = "insert"
            curr_atoms = self._insert(atoms, self.species, rng)
            self._extra_info = f"Insert_{self.species}"

        return curr_atoms

    def metropolis(self, prev_ene: float, curr_ene: float, rng=np.random) -> bool:
        """"""
        # - acceptance ratio
        kBT_eV = units.kB * self.temperature
        beta = 1.0 / kBT_eV  # 1/(kb*T), eV

        # - cubic thermo de broglie
        hplanck = units._hplanck  # J/Hz = kg*m2*s-1
        # _mass = np.sum([data.atomic_masses[data.atomic_numbers[e]] for e in expart]) # g/mol
        _species = convert_string_to_atoms(self.species)
        _species_mass = np.sum(_species.get_masses())
        # print("species mass: ", _mass)
        _mass = _species_mass * units._amu
        kbT_J = kBT_eV * units._e  # J = kg*m2*s-2
        cubic_wavelength = (
            hplanck / np.sqrt(2 * np.pi * _mass * kbT_J) * 1e10
        ) ** 3  # thermal de broglie wavelength

        # - compute coefficient
        # -- determine number of exchangeable particles
        # print("miaow: ", self._curr_tags_dict)
        if self.species not in self._curr_tags_dict:
            self._curr_tags_dict[self.species] = []
        nexatoms = len(self._curr_tags_dict[self.species])

        # -- compute composed coefficient
        if self._curr_operation == "insert":
            coef = self._curr_volume / (nexatoms + 1) / cubic_wavelength
            ene_diff = curr_ene - self.mu - prev_ene
        elif self._curr_operation == "remove":
            coef = nexatoms * cubic_wavelength / self._curr_volume
            ene_diff = curr_ene + self.mu - prev_ene
        else:
            raise RuntimeError(f"Unknown exchange operation {self._curr_operation}.")

        acc_ratio = np.min([1.0, coef * np.exp(-beta * (ene_diff))])

        content = "\nVolume %.4f Nexatoms %.4f CubicWave %.4f Coefficient %.4f\n" % (
            self._curr_volume,
            nexatoms,
            cubic_wavelength,
            coef,
        )
        # content = "\nVolume %.4f Beta %.4f Coefficient %.4f\n" %(
        #    0., beta, coef
        # )
        content += "Energy Difference %.4f [eV]\n" % ene_diff
        content += "Accept Ratio %.4f\n" % acc_ratio
        for x in content.split("\n"):
            self._print(x)

        rn_move = rng.uniform()
        self._print(f"{self.__class__.__name__} Probability %.4f" % rn_move)

        # - reset stored temp data
        self._curr_operation = None
        self._curr_tags_dict = None
        self._curr_volume = None

        return rn_move < acc_ratio

    def __repr__(self) -> str:
        """"""
        content = f"@Modifier {self.__class__.__name__}\n"
        content += (
            f"temperature {self.temperature} [K] pressure {self.pressure} [bar]\n"
        )
        content += "covalent ratio: \n"
        content += f"  min: {self.covalent_min} max: {self.covalent_max}\n"
        content += f"reservoir: "
        content += f"  species {self.species} with chemical potential {self.mu} [eV]\n"
        content += f"  within the region {self.region}\n"

        return content

    def as_dict(self) -> dict:
        """"""
        params = super().as_dict()
        params["reservoir"] = dict(species=self.species, mu=self.mu)
        params["use_bias"] = self.use_bias

        return params


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*

import os
import pathlib
from typing import NoReturn

from . import AbstractPotentialManager, DummyCalculator


class EamManager(AbstractPotentialManager):

    name = "eam"

    implemented_backends = ["lammps"]
    valid_combinations = (
        ("lammps", "lammps")
    )

    """See LAMMPS documentation for calculator parameters.
    """

    def __init__(self, *args, **kwargs):
        """"""
        super().__init__()

        return
    
    def register_calculator(self, calc_params, *agrs, **kwargs):
        """"""
        super().register_calculator(calc_params, *agrs, **kwargs)

        calc = DummyCalculator()

        # - some shared params
        command = calc_params.pop("command", None)
        directory = calc_params.pop("directory", pathlib.Path.cwd())

        type_list = calc_params.pop("type_list", [])
        type_map = {}
        for i, a in enumerate(type_list):
            type_map[a] = i
        
        # --- model files
        model_ = calc_params.get("model", [])
        if not isinstance(model_, list):
            model_ = [model_]

        models = []
        for m in model_:
            m = pathlib.Path(m).resolve()
            if not m.exists():
                raise FileNotFoundError(f"Cant find model file {str(m)}")
            models.append(str(m))
        
        if self.calc_backend == "lammps":
            from gdpx.computation.lammps import Lammps
            if models:
                pair_style = "eam"
                pair_coeff = calc_params.pop("pair_coeff", "* *")
                pair_coeff += " {} ".format(models[0])

                pair_style_name = pair_style.split()[0]
                assert pair_style_name == "eam", "Incorrect pair_style for lammps eam..."

                calc = Lammps(
                    command=command, directory=directory, 
                    pair_style=pair_style, pair_coeff=pair_coeff,
                    **calc_params
                )
                # - update several params
                calc.units = "metal"
                calc.atom_style = "atomic"
        else:
            ...
        
        self.calc = calc

        return


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import abc
import copy
import pathlib
from typing import NoReturn, Union, Callable

import numpy as np

from . import config

"""Every working component is represented by a node.
"""

class AbstractNode(abc.ABC):

    #: Node name.
    name: str = "node"

    #: The random seed when initialising the object.
    _init_random_seed: Union[int, dict] = None

    #: Working directory.
    _directory: pathlib.Path = "./"

    #: Standard print function.
    _print: Callable = config._print

    #: Standard debug function.
    _debug: Callable = config._debug

    #: Default parameters.
    default_parameters: dict = dict()

    def __init__(
        self, directory: Union[str,pathlib.Path]="./", 
        random_seed: Union[int, dict]=None, *args, **kwargs
    ):
        """"""
        # - set working directory
        self.directory = directory

        # - set random generator
        self._init_random_seed = random_seed
        self.set_rng(seed=random_seed)

        # - number of processors
        self.njobs = config.NJOBS

        return
    
    @property
    def init_random_seed(self):
        """The random seed at the initialisation of the object.

        This cannot be changed after the object is initialised.
        
        """

        return self._init_random_seed

    @property
    def directory(self) -> pathlib.Path:
        """Working directory.

        Note:
            When setting directory, some additional files are re-directed.

        """
        return self._directory
    
    @directory.setter
    def directory(self, directory_: Union[str,pathlib.Path]) -> NoReturn:
        self._directory = pathlib.Path(directory_)
        #if not self._directory.exists():
        #    self._directory.mkdir(parents=True)

        return 

    def set_rng(self, seed=None):
        """"""
        # - save random seed
        if seed is None:
            seed = int(config.GRNG.integers(0, 1e8))
        self.random_seed = seed

        # - assign random seeds
        if isinstance(seed, int): # TODO: numpy int?
            self.rng = np.random.Generator(np.random.PCG64(seed))
        elif isinstance(seed, dict): # TODO: omegaconf dict?
            self.rng = np.random.Generator(np.random.PCG64())
            self.rng.bit_generator.state = seed
        else:
            ...

        return


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import pathlib

import numpy as np

from ase import Atoms
from ase.io import read, write
from ase.calculators.singlepoint import SinglePointCalculator

"""Functions to convert different training datasets.
"""


def convert_dataset(pinp, format="dp"):
    """Convert dataset..."""
    pinp = pathlib.Path(pinp).resolve()
    print(pinp)

    # type.raw type_map.raw nopbc
    type_digits = np.loadtxt(pinp/"type.raw", dtype=int)
    type_list = np.loadtxt(pinp/"type_map.raw", dtype=str)
    chemical_symbols = [type_list[x] for x in type_digits]
    print(chemical_symbols)

    # box.npy  coord.npy  energy.npy  force.npy  virial.npy
    frames = []
    set_dirs = sorted(list(pinp.glob("set.*")))
    for p in set_dirs:
        box = np.load(p/"box.npy")
        coord = np.load(p/"coord.npy")
        energy = np.load(p/"energy.npy")
        force = np.load(p/"force.npy")
        if not (p/"virial.npy").exists():
            virial = None
        else:
            virial = np.load(p/"virial.npy") # unit: eV
        nframes = box.shape[0]
        curr_frames = []
        for i in range(nframes):
            atoms = Atoms(
                chemical_symbols, positions=coord[i].reshape(-1, 3), 
                cell=box[i].reshape(3, 3), pbc=True
            )
            results = dict(
                energy=energy[i], forces=force[i].reshape(-1, 3),
            )
            if virial is not None:
                stress = -0.5 * (virial[i] + virial[i].T) / atoms.get_volume()
                results["stress"] = stress[[0, 4, 8, 5, 2, 1]] # unit: eV/Ang^3
            calc = SinglePointCalculator(atoms, **results)
            atoms.calc = calc
            curr_frames.append(atoms)
        #write("./xxx.xyz", frames)
        frames.extend(curr_frames)
    write("./xxx.xyz", frames)

    return frames


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*

import re
import subprocess
import json
import pathlib

from .scheduler import AbstractScheduler


class PbsScheduler(AbstractScheduler):

    name = "pbs"

    PREFIX = "#"
    SUFFIX = ".pbs"
    SHELL = "#!/bin/bash -l"

    SUBMIT_COMMAND = "qsub"
    ENQUIRE_COMMAND = "qstat"

    default_parameters = {}

    running_status = ["R", "Q", "PD", "CG"]

    def __str__(self):
        """Return the content of the job script."""
        # - slurm params
        content = self.SHELL + "\n"
        for key, value in self.parameters.items():
            if value:
                content += "{} --{}={}\n".format(self.PREFIX, key, value)
            # else:
            #    raise ValueError("Keyword *%s* not properly set." %key)

        if self.environs:
            content += "\n\n"
            content += self.environs

        if self.user_commands:
            content += "\n\n"
            content += self.user_commands

        return content

    def is_finished(self) -> bool:
        """Not implemented yet."""

        raise NotImplementedError()


if __name__ == "__main__":
    ...
